@@ -1,20 +1,21 @@
 use super::{State, StateIdx, DFA};
 use crate::collections::Map;
-use crate::semantic_action_table::SemanticActionIdx;
+use crate::nfa::AcceptingState;
 
 #[derive(Debug)]
-pub enum Trans {
-    Accept(SemanticActionIdx),
+pub enum Trans<A> {
+    Accept(Vec<AcceptingState<A>>),
     Trans(StateIdx),
 }
 
 /// Removes accepting states with no transitions, makes the transitions to those states accepting.
-pub fn simplify<K>(
-    dfa: DFA<StateIdx, SemanticActionIdx>,
+pub fn simplify<K, A: Clone>(
+    dfa: DFA<StateIdx, A>,
     dfa_state_indices: &mut Map<K, StateIdx>,
-) -> DFA<Trans, SemanticActionIdx> {
-    let mut empty_states: Vec<(StateIdx, Option<SemanticActionIdx>)> = vec![];
-    let mut non_empty_states: Vec<(StateIdx, State<StateIdx, SemanticActionIdx>)> = vec![];
+) -> DFA<Trans<A>, A> {
+    let mut empty_states: Vec<(StateIdx, Vec<AcceptingState<A>>)> = vec![];
+
+    let mut non_empty_states: Vec<(StateIdx, State<StateIdx, A>)> = vec![];
 
     for (state_idx, state) in dfa.into_state_indices() {
         if state.has_no_transitions() && !state.initial {
@@ -31,14 +32,14 @@ pub fn simplify<K>(
         *t = t.map(|i| i - idx);
     }
 
-    let map_transition = |t: StateIdx| -> Option<Trans> {
+    let map_transition = |t: StateIdx| -> Trans<A> {
         match empty_states.binary_search_by(|(state_idx, _action)| state_idx.cmp(&t)) {
-            Ok(idx) => empty_states[idx].1.map(Trans::Accept),
-            Err(idx) => Some(Trans::Trans(t.map(|i| i - idx))),
+            Ok(idx) => Trans::Accept(empty_states[idx].1.clone()),
+            Err(idx) => Trans::Trans(t.map(|i| i - idx)),
         }
     };
 
-    let new_states: Vec<State<Trans, SemanticActionIdx>> = non_empty_states
+    let new_states: Vec<State<Trans<A>, A>> = non_empty_states
         .into_iter()
         .map(|(_state_idx, state)| {
             let State {
@@ -53,19 +54,19 @@ pub fn simplify<K>(
 
             let char_transitions = char_transitions
                 .into_iter()
-                .filter_map(|(char, next)| map_transition(next).map(|next| (char, next)))
+                .map(|(char, next)| (char, map_transition(next)))
                 .collect();
 
-            let range_transitions = range_transitions.filter_map(map_transition);
+            let range_transitions = range_transitions.map(map_transition);
 
-            let any_transition = any_transition.and_then(map_transition);
+            let any_transition = any_transition.map(map_transition);
 
-            let end_of_input_transition = end_of_input_transition.and_then(map_transition);
+            let end_of_input_transition = end_of_input_transition.map(map_transition);
 
             let predecessors = predecessors
                 .into_iter()
                 .map(|pred| match map_transition(pred) {
-                    Some(Trans::Trans(pred)) => pred,
+                    Trans::Trans(pred) => pred,
                     _ => {
                         // This pass should only remove nodes without successors, so it's a bug if
                         // we remove a predecessor
