@@ -2,6 +2,13 @@
 //!
 //! [1]: https://github.com/osa1/lexgen
 
+#![allow(
+    clippy::collapsible_else_if,
+    clippy::enum_variant_names,
+    clippy::too_many_arguments,
+    clippy::upper_case_acronyms
+)]
+
 mod ast;
 mod builtin;
 mod char_ranges;
@@ -12,16 +19,18 @@ mod nfa;
 mod nfa_to_dfa;
 mod range_map;
 mod regex_to_nfa;
+mod right_ctx;
 mod semantic_action_table;
 
 #[cfg(test)]
 mod tests;
 
-use ast::{Lexer, Regex, Rule, SingleRule, Var};
+use ast::{Lexer, Regex, RegexCtx, Rule, SingleRule, Var};
 use collections::Map;
 use dfa::{StateIdx as DfaStateIdx, DFA};
 use nfa::NFA;
 use nfa_to_dfa::nfa_to_dfa;
+use right_ctx::RightCtxDFAs;
 use semantic_action_table::{SemanticActionIdx, SemanticActionTable};
 
 use std::collections::hash_map::Entry;
@@ -47,6 +56,9 @@ pub fn lexer(input: TokenStream) -> TokenStream {
     // Maps DFA names to their initial states in the final DFA
     let mut dfas: Map<String, dfa::StateIdx> = Default::default();
 
+    // DFAs generated for right contexts
+    let mut right_ctx_dfas = RightCtxDFAs::new();
+
     let mut bindings: Map<Var, Regex> = Default::default();
 
     let mut dfa: Option<DFA<DfaStateIdx, SemanticActionIdx>> = None;
@@ -64,45 +76,25 @@ pub fn lexer(input: TokenStream) -> TokenStream {
                     panic!("Variable {:?} is defined multiple times", entry.key().0);
                 }
                 Entry::Vacant(entry) => {
-                    entry.insert(re);
+                    // TODO: Check that regex doesn't have right context
+                    entry.insert(re.re);
                 }
             },
             Rule::RuleSet { name, rules } => {
                 if name == "Init" {
-                    if dfa.is_some() {
-                        panic!("\"Init\" rule set can only be defined once");
-                    }
-                    let mut nfa: NFA<SemanticActionIdx> = NFA::new();
-                    for SingleRule { lhs, rhs } in rules {
-                        nfa.add_regex(&bindings, &lhs, rhs);
-                    }
+                    let dfa = dfa.insert(compile_rules(rules, &bindings, &mut right_ctx_dfas));
+                    let initial_state = dfa.initial_state();
 
-                    // println!("NFA=\n{}", nfa);
-
-                    let dfa_ = nfa_to_dfa(&nfa);
-                    let initial_state = dfa_.initial_state();
-
-                    // println!("DFA=\n{}", dfa_);
-
-                    dfa = Some(dfa_);
                     if dfas.insert(name.to_string(), initial_state).is_some() {
                         panic!("Rule set {:?} is defined multiple times", name.to_string());
                     }
                 } else {
-                    let dfa = match dfa.as_mut() {
-                        None => panic!("First rule set should be named \"Init\""),
-                        Some(dfa) => dfa,
-                    };
-                    let mut nfa: NFA<SemanticActionIdx> = NFA::new();
+                    let dfa = dfa
+                        .as_mut()
+                        .expect("First rule set should be named \"Init\"");
 
-                    for SingleRule { lhs, rhs } in rules {
-                        nfa.add_regex(&bindings, &lhs, rhs);
-                    }
+                    let dfa_ = compile_rules(rules, &bindings, &mut right_ctx_dfas);
 
-                    // println!("NFA=\n{}", nfa);
-
-                    let dfa_ = nfa_to_dfa(&nfa);
-                    // println!("DFA=\n{}", dfa_);
                     let dfa_idx = dfa.add_dfa(dfa_);
 
                     if dfas.insert(name.to_string(), dfa_idx).is_some() {
@@ -120,20 +112,8 @@ pub fn lexer(input: TokenStream) -> TokenStream {
                     );
                 }
 
-                let mut nfa: NFA<SemanticActionIdx> = NFA::new();
-                for SingleRule { lhs, rhs } in rules {
-                    nfa.add_regex(&bindings, &lhs, rhs);
-                }
-
-                // println!("NFA=\n{}", nfa);
-
-                let dfa_ = nfa_to_dfa(&nfa);
-
-                // println!("DFA=\n{}", dfa_);
-
-                let initial_state = dfa_.initial_state();
-
-                dfa = Some(dfa_);
+                let dfa = dfa.insert(compile_rules(rules, &bindings, &mut right_ctx_dfas));
+                let initial_state = dfa.initial_state();
                 dfas.insert("Init".to_owned(), initial_state);
             }
             Rule::ErrorType { ty } => match user_error_type {
@@ -145,8 +125,6 @@ pub fn lexer(input: TokenStream) -> TokenStream {
         }
     }
 
-    // println!("Final DFA=\n{}", dfa.as_ref().unwrap());
-
     // There should be a rule with name "Init"
     if dfas.get("Init").is_none() {
         panic!(
@@ -159,6 +137,7 @@ pub fn lexer(input: TokenStream) -> TokenStream {
 
     dfa::codegen::reify(
         dfa,
+        &right_ctx_dfas,
         semantic_action_table,
         user_state_type,
         user_error_type,
@@ -169,3 +148,23 @@ pub fn lexer(input: TokenStream) -> TokenStream {
     )
     .into()
 }
+
+fn compile_rules(
+    rules: Vec<SingleRule>,
+    bindings: &Map<Var, Regex>,
+    right_ctx_dfas: &mut RightCtxDFAs<DfaStateIdx>,
+) -> DFA<DfaStateIdx, SemanticActionIdx> {
+    let mut nfa: NFA<SemanticActionIdx> = NFA::new();
+
+    for SingleRule { lhs, rhs } in rules {
+        let RegexCtx { re, right_ctx } = lhs;
+
+        let right_ctx = right_ctx
+            .as_ref()
+            .map(|right_ctx| right_ctx_dfas.new_right_ctx(bindings, right_ctx));
+
+        nfa.add_regex(bindings, &re, right_ctx, rhs);
+    }
+
+    nfa_to_dfa(&nfa)
+}
