@@ -1,7 +1,7 @@
 mod test_utils;
 
 use lexgen::lexer;
-use lexgen_util::Loc;
+use lexgen_util::{LexerError, LexerErrorKind, Loc};
 use test_utils::{loc, next};
 
 use std::convert::TryFrom;
@@ -97,14 +97,13 @@ fn readme_2() {
 
         rule Count {
             '=' => |lexer| {
-                let n = *lexer.state();
-                *lexer.state() = n + 1;                     // line 16
-                lexer.continue_()                           // line 17
+                *lexer.state() += 1;                        // line 15
+                lexer.continue_()                           // line 16
             },
 
             '[' => |lexer| {
                 let n = *lexer.state();
-                lexer.switch_and_return(LexerRule::Init, n) // line 22
+                lexer.switch_and_return(LexerRule::Init, n) // line 21
             },
         }
     }
@@ -263,41 +262,6 @@ fn switch_user_state() {
     assert_eq!(lexer.next(), None);
 }
 
-#[test]
-fn counting() {
-    lexer! {
-        Lexer(usize) -> usize;
-
-        rule Init {
-            ' ',
-
-            '[' => |lexer| {
-                *lexer.state() = 0;
-                lexer.switch(LexerRule::Count)
-            },
-        }
-
-        rule Count {
-            '=' => |lexer| {
-                let n = *lexer.state();
-                *lexer.state() = n + 1;
-                lexer.continue_()
-            },
-
-            '[' => |lexer| {
-                let n = *lexer.state();
-                lexer.switch_and_return(LexerRule::Init, n)
-            },
-        }
-    }
-
-    let mut lexer = Lexer::new("[[ [=[ [==[");
-    assert_eq!(lexer.next(), Some(Ok((loc(0, 0, 0), 0, loc(0, 2, 2)))));
-    assert_eq!(lexer.next(), Some(Ok((loc(0, 3, 3), 1, loc(0, 6, 6)))));
-    assert_eq!(lexer.next(), Some(Ok((loc(0, 7, 7), 2, loc(0, 11, 11)))));
-    assert_eq!(lexer.next(), None);
-}
-
 #[test]
 fn lua_long_strings() {
     #[derive(Default, Clone, Copy)]
@@ -465,7 +429,10 @@ fn rule_kind_fallible_no_lifetimes() {
     );
     assert!(matches!(
         lexer.next(),
-        Some(Err(::lexgen_util::LexerError::Custom(_)))
+        Some(Err(LexerError {
+            kind: LexerErrorKind::Custom(_),
+            ..
+        }))
     ));
     assert_eq!(lexer.next(), None);
 }
@@ -502,7 +469,10 @@ fn rule_kind_fallible_with_lifetimes() {
     );
     assert!(matches!(
         lexer.next(),
-        Some(Err(::lexgen_util::LexerError::Custom(UserError("blah"))))
+        Some(Err(LexerError {
+            kind: LexerErrorKind::Custom(UserError("blah")),
+            ..
+        }))
     ));
     assert_eq!(lexer.next(), None);
 }
@@ -552,7 +522,10 @@ fn rule_kind_mix() {
     );
     assert!(matches!(
         lexer.next(),
-        Some(Err(::lexgen_util::LexerError::Custom(UserError("blah"))))
+        Some(Err(LexerError {
+            kind: LexerErrorKind::Custom(UserError("blah")),
+            ..
+        }))
     ));
     assert_eq!(lexer.next(), None);
 
@@ -991,3 +964,104 @@ fn loc_tracking() {
         Some(Ok((loc(1, 0, 17), "ｗｏｒｌｄ!!!", loc(1, 13, 35))))
     );
 }
+
+#[test]
+fn diff_1() {
+    lexer! {
+        Lexer -> &'input str;
+
+        let exclude = ['3'-'7'];
+
+        ['0'-'9'] # $exclude => |lexer| {
+            let match_ = lexer.match_();
+            lexer.return_(match_)
+        },
+    }
+
+    let mut lexer = Lexer::new("01289");
+    assert_eq!(next(&mut lexer), Some(Ok("0")));
+    assert_eq!(next(&mut lexer), Some(Ok("1")));
+    assert_eq!(next(&mut lexer), Some(Ok("2")));
+    assert_eq!(next(&mut lexer), Some(Ok("8")));
+    assert_eq!(next(&mut lexer), Some(Ok("9")));
+    assert_eq!(next(&mut lexer), None);
+
+    let mut lexer = Lexer::new("34567");
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+    assert!(matches!(next(&mut lexer), None));
+}
+
+#[test]
+fn diff_2() {
+    lexer! {
+        Lexer -> &'input str;
+
+        _ # 'a' => |lexer| {
+            let match_ = lexer.match_();
+            lexer.return_(match_)
+        },
+    }
+
+    let mut lexer = Lexer::new("b");
+    assert_eq!(next(&mut lexer), Some(Ok("b")));
+    assert!(matches!(next(&mut lexer), None));
+
+    let mut lexer = Lexer::new("a");
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+    assert!(matches!(next(&mut lexer), None));
+}
+
+#[test]
+fn diff_3() {
+    lexer! {
+        Lexer -> &'input str;
+
+        "'" (_ # ('\t' | '\n' | '\\' | '\'')) "'" => |lexer| {
+            let match_ = lexer.match_();
+            lexer.return_(match_)
+        },
+    }
+
+    let mut lexer = Lexer::new("''");
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+
+    let mut lexer = Lexer::new("'''");
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+
+    let mut lexer = Lexer::new("'\t'");
+    assert!(matches!(next(&mut lexer), Some(Err(_))));
+
+    let mut lexer = Lexer::new("'a'");
+    assert_eq!(next(&mut lexer), Some(Ok("'a'")));
+    assert_eq!(next(&mut lexer), None);
+}
+
+#[test]
+fn diff_4() {
+    lexer! {
+        Lexer -> &'input str;
+
+        "//" (_ # '\n')* $? => |lexer| {
+            let match_ = lexer.match_();
+            lexer.return_(match_)
+        },
+    }
+
+    let mut lexer = Lexer::new("// asdf");
+    assert_eq!(next(&mut lexer), Some(Ok("// asdf")));
+    assert_eq!(next(&mut lexer), None);
+
+    let mut lexer = Lexer::new("// asdf\n");
+    assert_eq!(next(&mut lexer), Some(Ok("// asdf")));
+    assert_eq!(
+        next(&mut lexer),
+        Some(Err(LexerError {
+            location: loc(0, 7, 7),
+            kind: LexerErrorKind::InvalidToken,
+        }))
+    );
+}
